package com.jdbernard.wdiwtlt

import com.jdbernard.wdiwtlt.db.ORM
import com.jdbernard.wdiwtlt.db.models.*

import org.jaudiotagger.audio.AudioFile
import org.jaudiotagger.audio.AudioFileIO
import org.jaudiotagger.tag.Tag as JATag
import org.jaudiotagger.tag.FieldKey
import org.slf4j.Logger
import org.slf4j.LoggerFactory
import org.apache.commons.codec.digest.DigestUtils

import static org.jaudiotagger.tag.FieldKey.*

public class MediaLibrary {

    private static Logger logger = LoggerFactory.getLogger(MediaLibrary)

    @Delegate ORM orm
    private File libraryRoot

    public MediaLibrary(ORM orm, File rootDir) {
        logger.debug("Creating a MediaLibrary rooted at: {}",
            rootDir.canonicalPath)

        this.orm = orm
        this.libraryRoot = rootDir }

    public void clean() {
        orm.removeEmptyAlbums()
        orm.removeEmptyArtists()
        orm.removeEmptyPlaylists()
    }

    public void rescanLibrary() { libraryRoot.eachFileRecurse { addFile(it) } }

    public MediaFile addFile(File f) {
        if (!f.exists() || !f.isFile()) {
            logger.info("Ignoring non-existant file: {}", f.canonicalPath)
            return null }

        def relPath = getRelativePath(libraryRoot, f)
        MediaFile mf = orm.getMediaFileByFilePath(relPath)

        if (mf) {
            logger.info(
                "Ignoring a media file I already know about: {}", relPath)
            return mf }

        // Read in the media's tags
        mf = new MediaFile()
        def af

        try { af = AudioFileIO.read(f) }
        catch (Exception e) {
            logger.info("Ignoring a file because I can't" +
                "read the media tag info:\n\t{}",
                e.localizedMessage)
            return null }

        def fileTag = af.tag

        mf.name = fileTag?.getFirst(TITLE)?.trim() ?: f.name
        mf.filePath = relPath
        mf.comment = fileTag?.getFirst(COMMENT)?.trim()
        mf.trackNumber = safeToInteger(fileTag?.getFirst(TRACK))

        def folderParts = mf.filePath.split("[\\\\/]")[1..<-1] as LinkedList

        // Find artist and album names (if any)
        def artistName = fileTag?.getFirst(ARTIST)?.trim()
        def albumName = fileTag?.getFirst(ALBUM)?.trim()

        if (!artistName) {
            mf.metaInfoSource = MediaFile.FILE_LOCATION
            artistName = folderParts.size() >= 2 ?  folderParts[0] : null }

        if (!albumName) {
            mf.metaInfoSource = MediaFile.FILE_LOCATION
            albumName = folderParts.peekLast() }

        // Hash the file
        mf.fileHash = f.withInputStream { DigestUtils.md5Hex(it) }

        orm.withTransaction {
            orm.create(mf)
            associateWithArtistAndAlbum(mf, artistName, albumName, fileTag)
        }
    }

    public def getByIdOrName(Class modelClass, String input) {
        def match
        if (safeToInteger(input))  match = [orm.getById(modelClass, safeToInteger(input))]
        else {
            match = orm.getByName(modelClass, input)
            if (!match) match = orm.getLike(modelClass, ["name"], [input]) }
        return match }

    public List<Artist> getArtistsByName(String name) {
        return orm.getArtistByName(name) ?: orm.getArtistsLikeName(name) }

    public List<Album> getAlbumsByName(String name) {
        return orm.getAlbumByName(name) ?: orm.getAlbumsLikeName(name) }

    private void associateWithArtistAndAlbum(MediaFile mf, String artistName,
    String albumName, JATag fileTag) {
        Artist artist = null
        Album album = null
        boolean newAlbumOrArtist = false

        if (artistName) {
            artist = orm.getArtistByName(artistName)
            if (!artist) {
                newAlbumOrArtist = true
                artist = new Artist(name: artistName)
                orm.create(artist) } }

        if (albumName) {
            album = orm.getAlbumByName(albumName)
            if (!album) {
                newAlbumOrArtist = true
                try {
                    album = new Album(name: albumName,
                        year: safeToInteger(fileTag?.getFirst(YEAR)),
                        trackTotal: safeToInteger(fileTag?.getFirst(TRACK_TOTAL))) }
                catch (UnsupportedOperationException use) {
                    album = new Album(name: albumName,
                        year: safeToInteger(fileTag?.getFirst(YEAR))) }
                orm.create(album) } }

        if (artist && album && newAlbumOrArtist)
            orm.addAlbumArtist(album.id, artist.id)

        if (artist) orm.associateMediaFileWithArtist(mf.id, artist.id)
        if (album) orm.associateMediaFileWithAlbum(mf.id, album.id)
    }

    /** #### `getRelativePath`
      * Given a parent path and a child path, assuming the child path is
      * contained within the parent path, return the relative path from the
      * parent to the child. */
    public static String getRelativePath(File parent, File child) {
        def parentPath = parent.canonicalPath.split("[\\\\/]")
        def childPath = child.canonicalPath.split("[\\\\/]")

        /// If the parent path is longer it cannot contain the child path and
        /// we cannot construct a relative path without backtracking.
        if (parentPath.length > childPath.length) return ""

        /// Compare the parent and child path up until the end of the parent
        /// path.
        int b = 0
        while (b < parentPath.length && parentPath[b] == childPath[b] ) b++

        /// If we stopped before reaching the end of the parent path it must be
        /// that the paths do not match. The parent cannot contain the child and
        /// we cannot build a relative path without backtracking.
        if (b != parentPath.length) return ""
        return (['.'] + childPath[b..<childPath.length]).join('/') }


    public static Integer safeToInteger(def val) {
        if (val == null) return null
        try { return val as Integer }
        catch (NumberFormatException nfe) { return null } }
}
